# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.process.properties import Interpolate
from buildbot.steps.configurable import BuildbotCiYml
from buildbot.steps.configurable import evaluate_condition
from buildbot.steps.configurable import parse_env_string
from buildbot.steps.shell import ShellCommand


class TestParseEnvString(unittest.TestCase):
    @parameterized.expand([
        ('single', 'ABC=1', {'ABC': '1'}),
        ('multiple', 'ABC=1 EFG=2', {'ABC': '1', 'EFG': '2'}),
        ('multiple_empty_quotes', 'ABC=\'\' EFG=2', {'ABC': '', 'EFG': '2'}),
        ('multiple_empty_double_quotes', 'ABC="" EFG=2', {'ABC': '', 'EFG': '2'}),
        ('multiple_single_quotes', 'ABC=\'1\' EFG=2', {'ABC': '1', 'EFG': '2'}),
        ('multiple_double_quotes', 'ABC="1" EFG=2', {'ABC': '1', 'EFG': '2'}),
        ('multiple_with_equals_in_value', 'ABC=1=2 EFG=2', {'ABC': '1=2', 'EFG': '2'}),
        (
            'multiple_with_equals_in_value_single_quotes',
            'ABC=\'1=2\' EFG=2',
            {'ABC': '1=2', 'EFG': '2'},
        ),
        (
            'multiple_with_equals_in_value_double_quotes',
            'ABC="1=2" EFG=2',
            {'ABC': '1=2', 'EFG': '2'},
        ),
        (
            'multiple_with_space_in_value_single_quotes',
            'ABC=\'1 2\' EFG=2',
            {'ABC': '1 2', 'EFG': '2'},
        ),
        (
            'multiple_with_space_in_value_double_quotes',
            'ABC="1 2" EFG=2',
            {'ABC': '1 2', 'EFG': '2'},
        ),
    ])
    def test_one(self, name, value, expected):
        self.assertEqual(parse_env_string(value), expected)

    @parameterized.expand([
        ('dot', 'ABC=1.0', {'ABC': '1.0'}),
        ('dash', 'ABC=1-0', {'ABC': '1-0'}),
        ('plus', 'ABC=1+0', {'ABC': '1+0'}),
        ('multiply', 'ABC=1*0', {'ABC': '1*0'}),
        ('tilde', 'ABC=1~0', {'ABC': '1~0'}),
        ('hat', 'ABC=1^0', {'ABC': '1^0'}),
        ('comma', 'ABC=1,0', {'ABC': '1,0'}),
        ('colon', 'ABC=1:0', {'ABC': '1:0'}),
        ('slash', 'ABC=1/0', {'ABC': '1/0'}),
        ('pipe', 'ABC=1|0', {'ABC': '1|0'}),
        ('excl_mark', 'ABC=1!0', {'ABC': '1!0'}),
        ('question_mark', 'ABC=1?0', {'ABC': '1?0'}),
        ('left_paren', 'ABC=1(0', {'ABC': '1(0'}),
        ('right_paren', 'ABC=1)0', {'ABC': '1)0'}),
        ('left_brace', 'ABC=1[0', {'ABC': '1[0'}),
        ('right_brace', 'ABC=1]0', {'ABC': '1]0'}),
        ('left_curly_brace', 'ABC=1{0', {'ABC': '1{0'}),
        ('right_curly_brace', 'ABC=1}0', {'ABC': '1}0'}),
        ('left_angled_brace', 'ABC=1<0', {'ABC': '1<0'}),
        ('right_angled_brace', 'ABC=1>0', {'ABC': '1>0'}),
    ])
    def test_special_characters(self, name, value, expected):
        self.assertEqual(parse_env_string(value), expected)

    def test_global_overridden(self):
        self.assertEqual(
            parse_env_string('K1=VE1 K2=VE2', {'K2': 'VG1', 'K3': 'VG3'}),
            {'K1': 'VE1', 'K2': 'VE2', 'K3': 'VG3'},
        )


class TestEvaluateCondition(unittest.TestCase):
    def test_bool(self):
        self.assertTrue(evaluate_condition('True', {}))
        self.assertFalse(evaluate_condition('False', {}))

    def test_string_empty(self):
        self.assertFalse(evaluate_condition('VALUE', {'VALUE': ''}))
        self.assertTrue(evaluate_condition('VALUE', {'VALUE': 'abc'}))

    def test_string_equal(self):
        self.assertTrue(evaluate_condition('VALUE == "a"', {'VALUE': 'a'}))
        self.assertFalse(evaluate_condition('VALUE == "a"', {'VALUE': 'b'}))

    def test_string_in_tuple(self):
        cond = 'VALUE in ("a", "b", "c", "d")'
        self.assertTrue(evaluate_condition(cond, {'VALUE': 'a'}))
        self.assertFalse(evaluate_condition(cond, {'VALUE': 'not'}))

    def test_string_not_in_tuple(self):
        cond = 'VALUE not in ("a", "b", "c", "d")'
        self.assertFalse(evaluate_condition(cond, {'VALUE': 'a'}))
        self.assertTrue(evaluate_condition(cond, {'VALUE': 'not'}))


class TestLoading(unittest.TestCase):
    def test_single_script(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - echo success
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': ['echo success'],
                'after_script': [],
            },
        )

    def test_single_script_interpolated_no_replacement(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - !i echo success
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': [Interpolate("echo success")],
                'after_script': [],
            },
        )

    def test_single_script_interpolated_with_replacement(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - !i echo "%(prop:name)s"
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': [Interpolate("echo %(prop:name)s")],
                'after_script': [],
            },
        )

    def test_single_script_dict_interpolate_with_replacement(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - title: mytitle
            - cmd: [ "echo", !i "%(prop:name)s" ]
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': [{'title': 'mytitle'}, {'cmd': ['echo', Interpolate('%(prop:name)s')]}],
                'after_script': [],
            },
        )

    def test_multiple_scripts(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - echo success
            - echo success2
            - echo success3
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': ['echo success', 'echo success2', 'echo success3'],
                'after_script': [],
            },
        )

    def test_script_with_step(self):
        c = BuildbotCiYml.load_from_str("""
        script:
            - !ShellCommand
              command: "echo success"
        """)
        self.assertEqual(
            c.script_commands,
            {
                'before_install': [],
                'install': [],
                'after_install': [],
                'before_script': [],
                'script': [ShellCommand(command='echo success')],
                'after_script': [],
            },
        )

    def test_matrix_include_simple(self):
        m = BuildbotCiYml.load_matrix(
            {'matrix': {'include': [{'env': 'ABC=10'}, {'env': 'ABC=11'}, {'env': 'ABC=12'}]}}, {}
        )
        self.assertEqual(
            m, [{'env': {'ABC': '10'}}, {'env': {'ABC': '11'}}, {'env': {'ABC': '12'}}]
        )

    def test_matrix_include_global(self):
        m = BuildbotCiYml.load_matrix(
            {'matrix': {'include': [{'env': 'ABC=10'}, {'env': 'ABC=11'}, {'env': 'ABC=12'}]}},
            {'GLOBAL': 'GV'},
        )
        self.assertEqual(
            m,
            [
                {'env': {'ABC': '10', 'GLOBAL': 'GV'}},
                {'env': {'ABC': '11', 'GLOBAL': 'GV'}},
                {'env': {'ABC': '12', 'GLOBAL': 'GV'}},
            ],
        )

    def test_matrix_include_global_with_override(self):
        m = BuildbotCiYml.load_matrix(
            {'matrix': {'include': [{'env': 'ABC=10'}, {'env': 'ABC=11'}, {'env': 'ABC=12'}]}},
            {'ABC': 'GV'},
        )
        self.assertEqual(
            m,
            [
                {'env': {'ABC': '10'}},
                {'env': {'ABC': '11'}},
                {'env': {'ABC': '12'}},
            ],
        )



#!/usr/bin/python
"""
svn.py
Script for BuildBot to monitor a remote Subversion repository.
Copyright (C) 2006 John Pye
"""
# This script is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os.path
import subprocess
import sys
import xml.dom.minidom

import ConfigParser

# change these settings to match your project
svnurl = "https://pse.cheme.cmu.edu/svn/ascend/code/trunk"
statefilename = "~/changemonitor/config.ini"
buildmaster = "buildbot.example.org:9989"  # connects to a PBChangeSource

xml1 = subprocess.check_output(
    "svn log --non-interactive --verbose --xml --limit=1 " + svnurl, shell=True
)
xml1 = xml1.decode(sys.stdout.encoding)
# print "XML\n-----------\n"+xml1+"\n\n"

try:
    doc = xml.dom.minidom.parseString(xml1)
    el = doc.getElementsByTagName("logentry")[0]
    revision = el.getAttribute("revision")
    author = "".join([t.data for t in el.getElementsByTagName("author")[0].childNodes])
    comments = "".join([t.data for t in el.getElementsByTagName("msg")[0].childNodes])

    pathlist = el.getElementsByTagName("paths")[0]
    paths = []
    for p in pathlist.getElementsByTagName("path"):
        paths.append("".join([t.data for t in p.childNodes]))
    # print "PATHS"
    # print paths
except xml.parsers.expat.ExpatError as e:
    print("FAILED TO PARSE 'svn log' XML:")
    print(str(e))
    print("----")
    print("RECEIVED TEXT:")
    print(xml1)
    import sys

    sys.exit(1)

fname = statefilename
fname = os.path.expanduser(fname)
ini = ConfigParser.SafeConfigParser()

try:
    ini.read(fname)
except Exception:
    print("Creating changemonitor config.ini:", fname)
    ini.add_section("CurrentRevision")
    ini.set("CurrentRevision", -1)

try:
    lastrevision = ini.get("CurrentRevision", "changeset")
except ConfigParser.NoOptionError:
    print("NO OPTION FOUND")
    lastrevision = -1
except ConfigParser.NoSectionError:
    print("NO SECTION FOUND")
    lastrevision = -1

if lastrevision != revision:
    # comments = codecs.encodings.unicode_escape.encode(comments)
    cmd = (
        "buildbot sendchange --master="
        + buildmaster
        + " --branch=trunk \
--revision=\""
        + revision
        + "\" --who=\""
        + author
        + "\" --vc=\"svn\" \
--comments=\""
        + comments
        + "\" "
        + " ".join(paths)
    )

    # print cmd
    res = subprocess.check_output(cmd)
    res = res.decode(sys.stdout.encoding)

    print("SUBMITTING NEW REVISION", revision)
    if not ini.has_section("CurrentRevision"):
        ini.add_section("CurrentRevision")
    try:
        ini.set("CurrentRevision", "changeset", revision)
        f = open(fname, "w")
        ini.write(f)
        # print "WROTE CHANGES TO",fname
    except Exception:
        print("FAILED TO RECORD INI FILE")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.schedulers import basic
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.state import StateTestMixin


class CommonStuffMixin:
    @defer.inlineCallbacks
    def makeScheduler(self, klass, **kwargs_override):
        kwargs = {"name": "tsched", "treeStableTimer": 60, "builderNames": ['tbuild']}
        kwargs.update(kwargs_override)

        yield self.master.db.insert_test_data([
            fakedb.Builder(name=builderName) for builderName in kwargs['builderNames']
        ])
        sched = yield self.attachScheduler(klass(**kwargs), self.OBJECTID, self.SCHEDULERID)

        # keep track of builds in self.events
        self.events = []

        @self.assertArgSpecMatches(sched.addBuildsetForChanges)
        def addBuildsetForChanges(
            waited_for=False,
            reason='',
            external_idstring=None,
            changeids=None,
            builderNames=None,
            properties=None,
            priority=None,
            **kw,
        ):
            self.assertEqual(external_idstring, None)
            self.assertEqual(reason, sched.reason)
            self.events.append(f"B{repr(changeids).replace(' ', '')}@{int(self.reactor.seconds())}")
            return defer.succeed(None)

        sched.addBuildsetForChanges = addBuildsetForChanges

        # see self.assertConsumingChanges
        self.consumingChanges = None

        def startConsumingChanges(**kwargs):
            self.consumingChanges = kwargs
            return defer.succeed(None)

        sched.startConsumingChanges = startConsumingChanges

        return sched

    def assertConsumingChanges(self, **kwargs):
        self.assertEqual(self.consumingChanges, kwargs)


class BaseBasicScheduler(
    CommonStuffMixin, scheduler.SchedulerMixin, StateTestMixin, TestReactorMixin, unittest.TestCase
):
    OBJECTID = 244
    SCHEDULERID = 4

    # a custom subclass since we're testing the base class.  This basically
    # re-implements SingleBranchScheduler, but with more asserts
    class Subclass(basic.BaseBasicScheduler):
        timer_started = False

        def getChangeFilter(self, *args, **kwargs):
            return kwargs.get('change_filter')

        def getTimerNameForChange(self, change):
            self.timer_started = True
            return "xxx"

        def getChangeClassificationsForTimer(self, sched_id, timer_name):
            assert timer_name == "xxx"
            assert sched_id == BaseBasicScheduler.SCHEDULERID
            return self.master.db.schedulers.getChangeClassifications(sched_id)

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    @defer.inlineCallbacks
    def mkch(self, **kwargs):
        # create changeset and insert in database.
        chd = {"branch": 'master', "project": '', "repository": ''}
        chd.update(kwargs)
        ch = self.makeFakeChange(**chd)
        # fakedb.Change requires changeid instead of number
        chd['changeid'] = chd['number']
        sourcestampid = chd['number'] + 100
        del chd['number']
        yield self.master.db.insert_test_data([
            fakedb.Change(sourcestampid=sourcestampid, **chd),
            fakedb.SourceStamp(id=sourcestampid),
        ])
        return ch

    # tests

    def test_constructor_positional_exception(self):
        with self.assertRaises(TypeError):
            self.Subclass("tsched", "master", 60)

    @defer.inlineCallbacks
    def test_activate_no_treeStableTimer(self):
        cf = mock.Mock('cf')
        fII = mock.Mock('fII')
        yield self.makeScheduler(
            self.Subclass, treeStableTimer=None, change_filter=cf, fileIsImportant=fII
        )

        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=92),
            fakedb.Change(changeid=20),
        ])

        yield self.master.db.schedulers.classifyChanges(self.SCHEDULERID, {20: True})

        yield self.master.startService()

        # check that the scheduler has started to consume changes, and the
        # classifications *have* been flushed, since they will not be used
        self.assertConsumingChanges(fileIsImportant=fII, change_filter=cf, onlyImportant=False)
        yield self.assert_classifications(self.SCHEDULERID, {})

    @defer.inlineCallbacks
    def test_activate_treeStableTimer(self):
        cf = mock.Mock()
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=10, change_filter=cf)

        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=92),
            fakedb.Change(changeid=20),
        ])
        yield self.master.db.schedulers.classifyChanges(self.SCHEDULERID, {20: True})

        yield self.master.startService()

        # check that the scheduler has started to consume changes, and no
        # classifications have been flushed.  Furthermore, the existing
        # classification should have been acted on, so the timer should be
        # running
        self.assertConsumingChanges(fileIsImportant=None, change_filter=cf, onlyImportant=False)
        yield self.assert_classifications(self.SCHEDULERID, {20: True})
        self.assertTrue(sched.timer_started)
        self.reactor.advance(10)

    @defer.inlineCallbacks
    def test_gotChange_no_treeStableTimer_unimportant(self):
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=None, branch='master')
        yield self.master.startService()

        yield sched.gotChange((yield self.mkch(branch='master', number=13)), False)

        self.assertEqual(self.events, [])

    @defer.inlineCallbacks
    def test_gotChange_no_treeStableTimer_important(self):
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=None, branch='master')

        yield self.master.startService()

        yield sched.gotChange((yield self.mkch(branch='master', number=13)), True)

        self.assertEqual(self.events, ['B[13]@0'])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_unimportant(self):
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=10, branch='master')
        yield self.master.startService()

        yield sched.gotChange((yield self.mkch(branch='master', number=13)), False)

        self.assertEqual(self.events, [])
        self.reactor.advance(10)
        self.assertEqual(self.events, [])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_important(self):
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=10, branch='master')

        yield self.master.startService()

        yield sched.gotChange((yield self.mkch(branch='master', number=13)), True)
        self.reactor.advance(10)

        self.assertEqual(self.events, ['B[13]@10'])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_sequence(self):
        sched = yield self.makeScheduler(self.Subclass, treeStableTimer=9, branch='master')
        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=92),
            fakedb.Change(changeid=1, branch='master', when_timestamp=1110),
            fakedb.ChangeFile(changeid=1, filename='readme.txt'),
            fakedb.Change(changeid=2, branch='master', when_timestamp=2220),
            fakedb.ChangeFile(changeid=2, filename='readme.txt'),
            fakedb.Change(changeid=3, branch='master', when_timestamp=3330),
            fakedb.ChangeFile(changeid=3, filename='readme.txt'),
            fakedb.Change(changeid=4, branch='master', when_timestamp=4440),
            fakedb.ChangeFile(changeid=4, filename='readme.txt'),
        ])
        yield self.master.startService()

        self.reactor.advance(2220)

        # this important change arrives at 2220, so the stable timer will last
        # until 2229
        yield sched.gotChange(self.makeFakeChange(branch='master', number=1, when=2220), True)
        self.assertEqual(self.events, [])
        yield self.assert_classifications(self.SCHEDULERID, {1: True})

        # but another (unimportant) change arrives before then
        self.reactor.advance(6)  # to 2226
        self.assertEqual(self.events, [])

        yield sched.gotChange(self.makeFakeChange(branch='master', number=2, when=2226), False)
        self.assertEqual(self.events, [])
        yield self.assert_classifications(self.SCHEDULERID, {1: True, 2: False})

        self.reactor.advance(3)  # to 2229
        self.assertEqual(self.events, [])

        self.reactor.advance(3)  # to 2232
        self.assertEqual(self.events, [])

        # another important change arrives at 2232
        yield sched.gotChange(self.makeFakeChange(branch='master', number=3, when=2232), True)
        self.assertEqual(self.events, [])
        yield self.assert_classifications(self.SCHEDULERID, {1: True, 2: False, 3: True})

        self.reactor.advance(3)  # to 2235
        self.assertEqual(self.events, [])

        # finally, time to start the build!
        self.reactor.advance(6)  # to 2241
        self.assertEqual(self.events, ['B[1,2,3]@2241'])
        yield self.assert_classifications(self.SCHEDULERID, {})

    @defer.inlineCallbacks
    def test_enabled_callback(self):
        sched = yield self.makeScheduler(self.Subclass)
        yield self.master.startService()

        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)

    @defer.inlineCallbacks
    def test_disabled_activate(self):
        sched = yield self.makeScheduler(self.Subclass)
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.activate()
        self.assertEqual(r, None)

    @defer.inlineCallbacks
    def test_disabled_deactivate(self):
        sched = yield self.makeScheduler(self.Subclass)
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.deactivate()
        self.assertEqual(r, None)


class SingleBranchScheduler(
    CommonStuffMixin, scheduler.SchedulerMixin, StateTestMixin, TestReactorMixin, unittest.TestCase
):
    SCHEDULERID = 245
    OBJECTID = 224455

    codebases = {
        'a': {'repository': "", 'branch': 'master'},
        'b': {'repository': "", 'branch': 'master'},
    }

    @defer.inlineCallbacks
    def makeFullScheduler(self, **kwargs):
        yield self.master.db.insert_test_data([
            fakedb.Builder(name=builderName) for builderName in kwargs['builderNames']
        ])
        sched = yield self.attachScheduler(
            basic.SingleBranchScheduler(**kwargs),
            self.OBJECTID,
            self.SCHEDULERID,
            overrideBuildsetMethods=True,
        )
        return sched

    def mkbs(self, **kwargs):
        # create buildset for expected_buildset in assertBuildset.
        bs = {
            "reason": self.sched.reason,
            "external_idstring": None,
            "sourcestampsetid": 100,
            "properties": [('scheduler', ('test', 'Scheduler'))],
        }
        bs.update(kwargs)
        return bs

    def mkss(self, **kwargs):
        # create sourcestamp for expected_sourcestamps in assertBuildset.
        ss = {"branch": 'master', "project": '', "repository": '', "sourcestampsetid": 100}
        ss.update(kwargs)
        return ss

    @defer.inlineCallbacks
    def mkch(self, **kwargs):
        # create changeset and insert in database.
        chd = {"branch": 'master', "project": '', "repository": ''}
        chd.update(kwargs)
        ch = self.makeFakeChange(**chd)
        # fakedb.Change requires changeid instead of number
        chd['changeid'] = chd['number']
        sourcestampid = chd['number'] + 100
        del chd['number']
        yield self.master.db.insert_test_data([
            fakedb.Change(sourcestampid=sourcestampid, **chd),
            fakedb.SourceStamp(id=sourcestampid),
        ])
        return ch

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    @defer.inlineCallbacks
    def test_constructor_no_reason(self):
        sched = yield self.makeScheduler(basic.SingleBranchScheduler, branch="master")
        yield sched.configureService()
        self.assertEqual(
            sched.reason, "The SingleBranchScheduler scheduler named 'tsched' triggered this build"
        )

    @defer.inlineCallbacks
    def test_constructor_reason(self):
        sched = yield self.makeScheduler(
            basic.SingleBranchScheduler, branch="master", reason="Changeset"
        )
        yield sched.configureService()
        self.assertEqual(sched.reason, "Changeset")

    def test_constructor_branch_mandatory(self):
        with self.assertRaises(config.ConfigErrors):
            basic.SingleBranchScheduler(name="tsched", treeStableTimer=60)

    def test_constructor_no_branch_but_filter(self):
        # this shouldn't fail
        basic.SingleBranchScheduler(
            name="tsched", treeStableTimer=60, builderNames=['a', 'b'], change_filter=mock.Mock()
        )

    def test_constructor_branches_forbidden(self):
        with self.assertRaises(config.ConfigErrors):
            basic.SingleBranchScheduler(name="tsched", treeStableTimer=60, branches='x')

    @defer.inlineCallbacks
    def test_constructor_priority_none(self):
        sched = yield self.makeScheduler(
            basic.SingleBranchScheduler, branch="master", priority=None
        )
        yield sched.configureService()
        self.assertEqual(sched.priority, None)

    @defer.inlineCallbacks
    def test_constructor_priority_int(self):
        sched = yield self.makeScheduler(basic.SingleBranchScheduler, branch="master", priority=8)
        yield sched.configureService()
        self.assertEqual(sched.priority, 8)

    @defer.inlineCallbacks
    def test_constructor_priority_function(self):
        def sched_priority(builderNames, changesByCodebase):
            return 0

        sched = yield self.makeScheduler(
            basic.SingleBranchScheduler, branch="master", priority=sched_priority
        )
        yield sched.configureService()
        self.assertEqual(sched.priority, sched_priority)

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_important(self):
        # this looks suspiciously like the same test above, because SingleBranchScheduler
        # is about the same as the test subclass used above
        sched = yield self.makeScheduler(
            basic.SingleBranchScheduler, treeStableTimer=10, branch='master'
        )

        yield self.master.startService()

        change = yield self.mkch(branch='master', number=13)
        yield sched.gotChange(change, True)
        self.reactor.advance(10)

        self.assertEqual(self.events, ['B[13]@10'])

    @defer.inlineCallbacks
    def test_gotChange_createAbsoluteSourceStamps_saveCodebase(self):
        # check codebase is stored after receiving change.
        sched = yield self.makeFullScheduler(
            name='test',
            builderNames=['test'],
            treeStableTimer=None,
            branch='master',
            codebases=self.codebases,
            createAbsoluteSourceStamps=True,
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.OBJECTID, name='test', class_name='SingleBranchScheduler')
        ])
        yield self.master.startService()

        yield sched.gotChange(
            (yield self.mkch(codebase='a', revision='1234:abc', repository='A', number=1)), True
        )
        yield sched.gotChange(
            (yield self.mkch(codebase='b', revision='2345:bcd', repository='B', number=2)), True
        )

        yield self.assert_state(
            self.OBJECTID,
            lastCodebases={
                'a': {
                    "branch": 'master',
                    "repository": 'A',
                    "revision": '1234:abc',
                    "lastChange": 1,
                },
                'b': {
                    "branch": 'master',
                    "repository": 'B',
                    "revision": '2345:bcd',
                    "lastChange": 2,
                },
            },
        )

    @defer.inlineCallbacks
    def test_gotChange_createAbsoluteSourceStamps_older_change(self):
        # check codebase is not stored if it's older than the most recent
        sched = yield self.makeFullScheduler(
            name='test',
            builderNames=['test'],
            treeStableTimer=None,
            branch='master',
            codebases=self.codebases,
            createAbsoluteSourceStamps=True,
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.OBJECTID, name='test', class_name='SingleBranchScheduler'),
            fakedb.ObjectState(
                objectid=self.OBJECTID,
                name='lastCodebases',
                value_json='{"a": {"branch": "master", "repository": "A", '
                '"revision": "5555:def",  "lastChange": 20}}',
            ),
        ])

        yield self.master.startService()

        # this change is not recorded, since it's older than
        # change 20
        yield sched.gotChange(
            (yield self.mkch(codebase='a', revision='1234:abc', repository='A', number=10)), True
        )

        yield self.assert_state(
            self.OBJECTID,
            lastCodebases={
                'a': {
                    "branch": 'master',
                    "repository": 'A',
                    "revision": '5555:def',
                    "lastChange": 20,
                }
            },
        )

    @defer.inlineCallbacks
    def test_getCodebaseDict(self):
        sched = yield self.makeFullScheduler(
            name='test',
            builderNames=['test'],
            treeStableTimer=None,
            branch='master',
            codebases=self.codebases,
            createAbsoluteSourceStamps=True,
        )
        yield self.master.startService()
        sched._lastCodebases = {
            'a': {"branch": 'master', "repository": 'A', "revision": '5555:def', "lastChange": 20}
        }

        cbd = yield sched.getCodebaseDict('a')
        self.assertEqual(
            cbd, {"branch": 'master', "repository": 'A', "revision": '5555:def', "lastChange": 20}
        )

    @defer.inlineCallbacks
    def test_getCodebaseDict_no_createAbsoluteSourceStamps(self):
        sched = yield self.makeFullScheduler(
            name='test',
            builderNames=['test'],
            treeStableTimer=None,
            branch='master',
            codebases=self.codebases,
            createAbsoluteSourceStamps=False,
        )
        yield self.master.startService()
        sched._lastCodebases = {
            'a': {"branch": 'master', "repository": 'A', "revision": '5555:def', "lastChange": 20}
        }

        cbd = yield sched.getCodebaseDict('a')
        # _lastCodebases is ignored
        self.assertEqual(cbd, {'branch': 'master', 'repository': ''})

    @defer.inlineCallbacks
    def test_gotChange_with_priority(self):
        sched = yield self.makeFullScheduler(
            name='test', builderNames=['test'], branch='master', priority=8
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.OBJECTID, name='test', class_name='SingleBranchScheduler')
        ])

        yield self.master.startService()

        yield sched.gotChange(
            (yield self.mkch(codebase='a', revision='1234:abc', repository='A', number=10)), True
        )

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForChanges',
                    {
                        'waited_for': False,
                        'external_idstring': None,
                        'changeids': [10],
                        'properties': None,
                        'reason': "The SingleBranchScheduler scheduler named 'test' triggered this build",
                        'builderNames': None,
                        'priority': 8,
                    },
                )
            ],
        )


class AnyBranchScheduler(
    CommonStuffMixin, scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase
):
    SCHEDULERID = 6
    OBJECTID = 246

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    def test_constructor_branch_forbidden(self):
        with self.assertRaises(config.ConfigErrors):
            basic.SingleBranchScheduler(name="tsched", treeStableTimer=60, branch='x')

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_multiple_branches(self):
        """Two changes with different branches get different treeStableTimers"""
        sched = yield self.makeScheduler(
            basic.AnyBranchScheduler, treeStableTimer=10, branches=['master', 'devel', 'boring']
        )

        yield self.master.startService()

        @defer.inlineCallbacks
        def mkch(**kwargs):
            ch = self.makeFakeChange(**kwargs)
            ch = yield self.addFakeChange(ch)
            return ch

        yield sched.gotChange((yield mkch(branch='master', number=500)), True)
        yield self.reactor.advance(1)  # time is now 1
        yield sched.gotChange((yield mkch(branch='master', number=501)), False)
        yield sched.gotChange((yield mkch(branch='boring', number=502)), False)
        yield self.reactor.pump([1] * 4)  # time is now 5
        yield sched.gotChange((yield mkch(branch='devel', number=503)), True)
        yield self.reactor.pump([1] * 10)  # time is now 15

        self.assertEqual(self.events, ['B[500,501]@11', 'B[503]@15'])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_multiple_repositories(self):
        """Two repositories, even with the same branch name, have different treeStableTimers"""
        sched = yield self.makeScheduler(
            basic.AnyBranchScheduler, treeStableTimer=10, branches=['master']
        )

        yield self.master.startService()

        @defer.inlineCallbacks
        def mkch(**kwargs):
            ch = self.makeFakeChange(**kwargs)
            ch = yield self.addFakeChange(ch)
            return ch

        yield sched.gotChange((yield mkch(branch='master', repository="repo", number=500)), True)
        yield self.reactor.advance(1)  # time is now 1
        yield sched.gotChange((yield mkch(branch='master', repository="repo", number=501)), False)
        yield sched.gotChange(
            (yield mkch(branch='master', repository="other_repo", number=502)), False
        )
        yield self.reactor.pump([1] * 4)  # time is now 5
        yield sched.gotChange(
            (yield mkch(branch='master', repository="other_repo", number=503)), True
        )
        yield self.reactor.pump([1] * 10)  # time is now 15

        self.assertEqual(self.events, ['B[500,501]@11', 'B[502,503]@15'])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_multiple_projects(self):
        """Two projects, even with the same branch name, have different treeStableTimers"""
        sched = yield self.makeScheduler(
            basic.AnyBranchScheduler, treeStableTimer=10, branches=['master']
        )

        yield self.master.startService()

        @defer.inlineCallbacks
        def mkch(**kwargs):
            ch = self.makeFakeChange(**kwargs)
            ch = yield self.addFakeChange(ch)
            return ch

        yield sched.gotChange((yield mkch(branch='master', project="proj", number=500)), True)
        yield self.reactor.advance(1)  # time is now 1
        yield sched.gotChange((yield mkch(branch='master', project="proj", number=501)), False)
        yield sched.gotChange(
            (yield mkch(branch='master', project="other_proj", number=502)), False
        )
        yield self.reactor.pump([1] * 4)  # time is now 5
        yield sched.gotChange((yield mkch(branch='master', project="other_proj", number=503)), True)
        yield self.reactor.pump([1] * 10)  # time is now 15

        self.assertEqual(self.events, ['B[500,501]@11', 'B[502,503]@15'])

    @defer.inlineCallbacks
    def test_gotChange_treeStableTimer_multiple_codebases(self):
        """Two codebases, even with the same branch name, have different treeStableTimers"""
        sched = yield self.makeScheduler(
            basic.AnyBranchScheduler, treeStableTimer=10, branches=['master']
        )

        yield self.master.startService()

        @defer.inlineCallbacks
        def mkch(**kwargs):
            ch = self.makeFakeChange(**kwargs)
            ch = yield self.addFakeChange(ch)
            return ch

        yield sched.gotChange((yield mkch(branch='master', codebase="base", number=500)), True)
        self.reactor.advance(1)  # time is now 1
        yield sched.gotChange((yield mkch(branch='master', codebase="base", number=501)), False)
        yield sched.gotChange(
            (yield mkch(branch='master', codebase="other_base", number=502)), False
        )
        self.reactor.pump([1] * 4)  # time is now 5
        yield sched.gotChange(
            (yield mkch(branch='master', codebase="other_base", number=503)), True
        )
        self.reactor.pump([1] * 10)  # time is now 15

        self.assertEqual(self.events, ['B[500,501]@11', 'B[502,503]@15'])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from twisted.trial import unittest

from buildbot_worker.scripts import create_worker
from buildbot_worker.test.util import misc

try:
    from unittest import mock
except ImportError:
    from unittest import mock

if TYPE_CHECKING:
    from typing import Any
    from typing import Literal
    from typing import Mapping

    from buildbot_worker.scripts.base import Options


def _regexp_path(name: str, *names: str) -> str:
    """
    Join two or more path components and create a regexp that will match that
    path.
    """
    return os.path.join(name, *names).replace("\\", "\\\\")


class TestDefaultOptionsMixin:
    # default options and required arguments
    options: Options = {
        # flags
        "no-logrotate": False,
        "relocatable": False,
        "quiet": False,
        "use-tls": False,
        "delete-leftover-dirs": False,
        # options
        "basedir": "bdir",
        "allow-shutdown": None,
        "umask": None,
        "log-size": 16,
        "log-count": 8,
        "keepalive": 4,
        "maxdelay": 2,
        "numcpus": None,
        "protocol": "pb",
        "maxretries": None,
        "connection-string": None,
        "proxy-connection-string": None,
        # arguments
        "host": "masterhost",
        "port": 1234,
        "name": "workername",
        "passwd": "orange",
    }


class TestMakeTAC(TestDefaultOptionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.create_worker._make_tac()
    """

    def assert_tac_file_contents(
        self,
        tac_contents: str,
        expected_args: Mapping[str, Any],
        relocate: str | None = None,
    ) -> None:
        """
        Check that generated TAC file is a valid Python script and it does what
        is typical for TAC file logic. Mainly create instance of Worker with
        expected arguments.
        """

        # pylint: disable=import-outside-toplevel
        # import modules for mocking
        import twisted.application.service
        import twisted.python.logfile

        import buildbot_worker.bot

        # mock service.Application class
        application_mock = mock.Mock()
        application_class_mock = mock.Mock(return_value=application_mock)
        self.patch(twisted.application.service, "Application", application_class_mock)

        # mock logging stuff
        logfile_mock = mock.Mock()
        self.patch(twisted.python.logfile.LogFile, "fromFullPath", logfile_mock)

        # mock Worker class
        worker_mock = mock.Mock()
        worker_class_mock = mock.Mock(return_value=worker_mock)
        self.patch(buildbot_worker.bot, "Worker", worker_class_mock)

        # Executed .tac file with mocked functions with side effect.
        # This will raise exception if .tac file is not valid Python file.
        globals_dict = {}
        if relocate:
            globals_dict["__file__"] = os.path.join(relocate, "buildbot.tac")
        exec(tac_contents, globals_dict, globals_dict)  # pylint: disable=exec-used

        # only one Application must be created in .tac
        application_class_mock.assert_called_once_with("buildbot-worker")

        # check that Worker created with passed options
        worker_class_mock.assert_called_once_with(
            expected_args["host"],
            expected_args["port"],
            expected_args["name"],
            expected_args["passwd"],
            expected_args["basedir"],
            expected_args["keepalive"],
            umask=expected_args["umask"],
            numcpus=expected_args["numcpus"],
            protocol=expected_args["protocol"],
            maxdelay=expected_args["maxdelay"],
            allow_shutdown=expected_args["allow-shutdown"],
            maxRetries=expected_args["maxretries"],
            useTls=expected_args["use-tls"],
            delete_leftover_dirs=expected_args["delete-leftover-dirs"],
            connection_string=expected_args["connection-string"],
            proxy_connection_string=expected_args["proxy-connection-string"],
        )

        # check that Worker instance attached to application
        self.assertEqual(worker_mock.method_calls, [mock.call.setServiceParent(application_mock)])

        # .tac file must define global variable "application", instance of
        # Application
        self.assertTrue(
            'application' in globals_dict, ".tac file doesn't define \"application\" variable"
        )
        self.assertTrue(
            globals_dict['application'] is application_mock,
            "defined \"application\" variable in .tac file is not Application instance",
        )

    def test_default_tac_contents(self) -> None:
        """
        test that with default options generated TAC file is valid.
        """
        tac_contents = create_worker._make_tac(self.options.copy())

        self.assert_tac_file_contents(tac_contents, self.options)

    def test_backslash_in_basedir(self) -> None:
        """
        test that using backslash (typical for Windows platform) in basedir
        won't break generated TAC file.
        """
        options = self.options.copy()
        options["basedir"] = r"C:\buildbot-worker dir\\"

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options)

    def test_quotes_in_basedir(self) -> None:
        """
        test that using quotes in basedir won't break generated TAC file.
        """
        options = self.options.copy()
        options["basedir"] = r"Buildbot's \"dir"

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options)

    def test_double_quotes_in_basedir(self) -> None:
        """
        test that using double quotes at begin and end of basedir won't break
        generated TAC file.
        """
        options = self.options.copy()
        options["basedir"] = r"\"\"Buildbot''"

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options)

    def test_special_characters_in_options(self) -> None:
        """
        test that using special characters in options strings won't break
        generated TAC file.
        """
        test_string = "\"\" & | ^ # @ \\& \\| \\^ \\# \\@ \\n \x07 \" \\\" ' \\' ''"
        options = self.options.copy()
        options["basedir"] = test_string
        options["host"] = test_string
        options["passwd"] = test_string
        options["name"] = test_string

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options)

    def test_flags_with_non_default_values(self) -> None:
        """
        test that flags with non-default values will be correctly written to
        generated TAC file.
        """
        options = self.options.copy()
        options["quiet"] = True
        options["use-tls"] = True
        options["delete-leftover-dirs"] = True

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options)

    def test_log_rotate(self) -> None:
        """
        test that when --no-logrotate options is not used, correct tac file
        is generated.
        """
        options = self.options.copy()
        options["no-logrotate"] = False

        tac_contents = create_worker._make_tac(options.copy())

        self.assertIn("from twisted.python.logfile import LogFile", tac_contents)
        self.assert_tac_file_contents(tac_contents, options)

    def test_no_log_rotate(self) -> None:
        """
        test that when --no-logrotate options is used, correct tac file
        is generated.
        """
        options = self.options.copy()
        options["no-logrotate"] = True

        tac_contents = create_worker._make_tac(options.copy())

        self.assertNotIn("from twisted.python.logfile import LogFile", tac_contents)
        self.assert_tac_file_contents(tac_contents, options)

    def test_relocatable_true(self) -> None:
        """
        test that when --relocatable option is True, worker is created from
        generated TAC file with correct basedir argument before and after
        relocation.
        """
        options = self.options.copy()
        options["relocatable"] = True
        options["basedir"] = os.path.join(os.getcwd(), "worker1")

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options, relocate=options["basedir"])

        _relocate = os.path.join(os.getcwd(), "worker2")
        options["basedir"] = _relocate
        self.assert_tac_file_contents(tac_contents, options, relocate=_relocate)

    def test_relocatable_false(self) -> None:
        """
        test that when --relocatable option is False, worker is created from
        generated TAC file with the same basedir argument before and after
        relocation.
        """
        options = self.options.copy()
        options["relocatable"] = False
        options["basedir"] = os.path.join(os.getcwd(), "worker1")

        tac_contents = create_worker._make_tac(options.copy())

        self.assert_tac_file_contents(tac_contents, options, relocate=options["basedir"])

        _relocate = os.path.join(os.getcwd(), "worker2")
        self.assert_tac_file_contents(tac_contents, options, relocate=_relocate)

    def test_options_with_non_default_values(self) -> None:
        """
        test that options with non-default values will be correctly written to
        generated TAC file and used as argument of Worker.
        """
        options = self.options.copy()
        options["allow-shutdown"] = "signal"
        options["umask"] = "18"
        options["log-size"] = 160
        options["log-count"] = "80"
        options["keepalive"] = 40
        options["maxdelay"] = 20
        options["numcpus"] = "10"
        options["protocol"] = "null"
        options["maxretries"] = "1"
        options["proxy-connection-string"] = "TCP:proxy.com:8080"

        tac_contents = create_worker._make_tac(options.copy())

        # These values are expected to be used as non-string literals in
        # generated TAC file.
        self.assertIn("rotateLength = 160", tac_contents)
        self.assertIn("maxRotatedFiles = 80", tac_contents)
        self.assertIn("keepalive = 40", tac_contents)
        self.assertIn("maxdelay = 20", tac_contents)
        self.assertIn("umask = 18", tac_contents)
        self.assertIn("numcpus = 10", tac_contents)
        self.assertIn("maxretries = 1", tac_contents)

        # Check also as arguments used in Worker initialization.
        options["umask"] = 18
        options["numcpus"] = 10
        options["maxretries"] = 1
        self.assert_tac_file_contents(tac_contents, options)

    def test_umask_octal_value(self) -> None:
        """
        test that option umask with octal value will be correctly written to
        generated TAC file and used as argument of Worker.
        """
        options = self.options.copy()
        options["umask"] = "0o22"

        tac_contents = create_worker._make_tac(options.copy())

        self.assertIn("umask = 0o22", tac_contents)
        options["umask"] = 18
        self.assert_tac_file_contents(tac_contents, options)

    def test_connection_string(self) -> None:
        """
        test that when --connection-string options is used, correct tac file
        is generated.
        """
        options = self.options.copy()
        options["connection-string"] = "TLS:buildbot-master.com:9989"

        tac_contents = create_worker._make_tac(options.copy())

        options["host"] = None
        options["port"] = None
        self.assert_tac_file_contents(tac_contents, options)


class TestMakeBaseDir(misc.StdoutAssertionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.create_worker._makeBaseDir()
    """

    def setUp(self) -> None:
        # capture stdout
        self.setUpStdoutAssertions()

        # patch os.mkdir() to do nothing
        self.mkdir = mock.Mock()
        self.patch(os, "mkdir", self.mkdir)

    def testBasedirExists(self) -> None:
        """
        test calling _makeBaseDir() on existing base directory
        """
        self.patch(os.path, "exists", mock.Mock(return_value=True))

        # call _makeBaseDir()
        create_worker._makeBaseDir("dummy", False)

        # check that correct message was printed to stdout
        self.assertStdoutEqual("updating existing installation\n")
        # check that os.mkdir was not called
        self.assertFalse(self.mkdir.called, "unexpected call to os.mkdir()")

    def testBasedirExistsQuiet(self) -> None:
        """
        test calling _makeBaseDir() on existing base directory with
        quiet flag enabled
        """
        self.patch(os.path, "exists", mock.Mock(return_value=True))

        # call _makeBaseDir()
        create_worker._makeBaseDir("dummy", True)

        # check that nothing was printed to stdout
        self.assertWasQuiet()
        # check that os.mkdir was not called
        self.assertFalse(self.mkdir.called, "unexpected call to os.mkdir()")

    def testBasedirCreated(self) -> None:
        """
        test creating new base directory with _makeBaseDir()
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))

        # call _makeBaseDir()
        create_worker._makeBaseDir("dummy", False)

        # check that os.mkdir() was called with correct path
        self.mkdir.assert_called_once_with("dummy")
        # check that correct message was printed to stdout
        self.assertStdoutEqual("mkdir dummy\n")

    def testBasedirCreatedQuiet(self) -> None:
        """
        test creating new base directory with _makeBaseDir()
        and quiet flag enabled
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))

        # call _makeBaseDir()
        create_worker._makeBaseDir("dummy", True)

        # check that os.mkdir() was called with correct path
        self.mkdir.assert_called_once_with("dummy")
        # check that nothing was printed to stdout
        self.assertWasQuiet()

    def testMkdirError(self) -> None:
        """
        test that _makeBaseDir() handles error creating directory correctly
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))

        # patch os.mkdir() to raise an exception
        self.patch(os, "mkdir", mock.Mock(side_effect=OSError(0, "dummy-error")))

        # check that correct exception was raised
        with self.assertRaisesRegex(
            create_worker.CreateWorkerError, "error creating directory dummy"
        ):
            create_worker._makeBaseDir("dummy", False)


class TestMakeBuildbotTac(misc.StdoutAssertionsMixin, misc.FileIOMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.create_worker._makeBuildbotTac()
    """

    def setUp(self) -> None:
        # capture stdout
        self.setUpStdoutAssertions()

        # patch os.chmod() to do nothing
        self.chmod = mock.Mock()
        self.patch(os, "chmod", self.chmod)

        # generate OS specific relative path to buildbot.tac inside basedir
        self.tac_file_path = _regexp_path("bdir", "buildbot.tac")

    def testTacOpenError(self) -> None:
        """
        test that _makeBuildbotTac() handles open() errors on buildbot.tac
        """
        self.patch(os.path, "exists", mock.Mock(return_value=True))
        # patch open() to raise exception
        self.setUpOpenError()

        # call _makeBuildbotTac() and check that correct exception is raised
        expected_message = f"error reading {self.tac_file_path}"
        with self.assertRaisesRegex(create_worker.CreateWorkerError, expected_message):
            create_worker._makeBuildbotTac("bdir", "contents", False)

    def testTacReadError(self) -> None:
        """
        test that _makeBuildbotTac() handles read() errors on buildbot.tac
        """
        self.patch(os.path, "exists", mock.Mock(return_value=True))
        # patch read() to raise exception
        self.setUpReadError()

        # call _makeBuildbotTac() and check that correct exception is raised
        expected_message = f"error reading {self.tac_file_path}"
        with self.assertRaisesRegex(create_worker.CreateWorkerError, expected_message):
            create_worker._makeBuildbotTac("bdir", "contents", False)

    def testTacWriteError(self) -> None:
        """
        test that _makeBuildbotTac() handles write() errors on buildbot.tac
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))
        # patch write() to raise exception
        self.setUpWriteError(0)

        # call _makeBuildbotTac() and check that correct exception is raised
        expected_message = f"could not write {self.tac_file_path}"
        with self.assertRaisesRegex(create_worker.CreateWorkerError, expected_message):
            create_worker._makeBuildbotTac("bdir", "contents", False)

    def checkTacFileCorrect(self, quiet: bool) -> None:
        """
        Utility function to test calling _makeBuildbotTac() on base directory
        with existing buildbot.tac file, which does not need to be changed.

        @param quiet: the value of 'quiet' argument for _makeBuildbotTac()
        """
        # set-up mocks to simulate buildbot.tac file in the basedir
        self.patch(os.path, "exists", mock.Mock(return_value=True))
        self.setUpOpen("test-tac-contents")

        # call _makeBuildbotTac()
        create_worker._makeBuildbotTac("bdir", "test-tac-contents", quiet)

        # check that write() was not called
        self.assertFalse(self.fileobj.write.called, "unexpected write() call")

        # check output to stdout
        if quiet:
            self.assertWasQuiet()
        else:
            self.assertStdoutEqual("buildbot.tac already exists and is correct\n")

    def testTacFileCorrect(self) -> None:
        """
        call _makeBuildbotTac() on base directory which contains a buildbot.tac
        file, which does not need to be changed
        """
        self.checkTacFileCorrect(False)

    def testTacFileCorrectQuiet(self) -> None:
        """
        call _makeBuildbotTac() on base directory which contains a buildbot.tac
        file, which does not need to be changed. Check that quite flag works
        """
        self.checkTacFileCorrect(True)

    def checkDiffTacFile(self, quiet: bool) -> None:
        """
        Utility function to test calling _makeBuildbotTac() on base directory
        with a buildbot.tac file, with does needs to be changed.

        @param quiet: the value of 'quiet' argument for _makeBuildbotTac()
        """
        # set-up mocks to simulate buildbot.tac file in basedir
        self.patch(os.path, "exists", mock.Mock(return_value=True))
        self.setUpOpen("old-tac-contents")

        # call _makeBuildbotTac()
        create_worker._makeBuildbotTac("bdir", "new-tac-contents", quiet)

        # check that buildbot.tac.new file was created with expected contents
        tac_file_path = os.path.join("bdir", "buildbot.tac")
        self.open.assert_has_calls([
            mock.call(tac_file_path),
            mock.call(tac_file_path + ".new", "w"),
        ])
        self.fileobj.write.assert_called_once_with("new-tac-contents")
        self.chmod.assert_called_once_with(tac_file_path + ".new", 0o600)

        # check output to stdout
        if quiet:
            self.assertWasQuiet()
        else:
            self.assertStdoutEqual(
                "not touching existing buildbot.tac\ncreating buildbot.tac.new instead\n"
            )

    def testDiffTacFile(self) -> None:
        """
        call _makeBuildbotTac() on base directory which contains a buildbot.tac
        file, with does needs to be changed.
        """
        self.checkDiffTacFile(False)

    def testDiffTacFileQuiet(self) -> None:
        """
        call _makeBuildbotTac() on base directory which contains a buildbot.tac
        file, with does needs to be changed. Check that quite flag works
        """
        self.checkDiffTacFile(True)

    def testNoTacFile(self) -> None:
        """
        call _makeBuildbotTac() on base directory with no buildbot.tac file
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))
        # capture calls to open() and write()
        self.setUpOpen()

        # call _makeBuildbotTac()
        create_worker._makeBuildbotTac("bdir", "test-tac-contents", False)

        # check that buildbot.tac file was created with expected contents
        tac_file_path = os.path.join("bdir", "buildbot.tac")
        self.open.assert_called_once_with(tac_file_path, "w")
        self.fileobj.write.assert_called_once_with("test-tac-contents")
        self.chmod.assert_called_once_with(tac_file_path, 0o600)


class TestMakeInfoFiles(misc.StdoutAssertionsMixin, misc.FileIOMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.create_worker._makeInfoFiles()
    """

    def setUp(self) -> None:
        # capture stdout
        self.setUpStdoutAssertions()

    def checkMkdirError(self, quiet: bool) -> None:
        """
        Utility function to test _makeInfoFiles() when os.mkdir() fails.

        Patch os.mkdir() to raise an exception, and check that _makeInfoFiles()
        handles mkdir errors correctly.

        @param quiet: the value of 'quiet' argument for _makeInfoFiles()
        """
        self.patch(os.path, "exists", mock.Mock(return_value=False))
        # patch os.mkdir() to raise an exception
        self.patch(os, "mkdir", mock.Mock(side_effect=OSError(0, "err-msg")))

        # call _makeInfoFiles() and check that correct exception is raised
        with self.assertRaisesRegex(
            create_worker.CreateWorkerError,
            "error creating directory {}".format(_regexp_path("bdir", "info")),
        ):
            create_worker._makeInfoFiles("bdir", quiet)

        # check output to stdout
        if quiet:
            self.assertWasQuiet()
        else:
            self.assertStdoutEqual("mkdir {}\n".format(os.path.join("bdir", "info")))

    def testMkdirError(self) -> None:
        """
        test _makeInfoFiles() when os.mkdir() fails
        """
        self.checkMkdirError(False)

    def testMkdirErrorQuiet(self) -> None:
        """
        test _makeInfoFiles() when os.mkdir() fails and quiet flag is enabled
        """
        self.checkMkdirError(True)

    def checkIOError(self, error_type: Literal["open", "write"], quiet: bool) -> None:
        """
        Utility function to test _makeInfoFiles() when open() or write() fails.

        Patch file IO functions to raise an exception, and check that
        _makeInfoFiles() handles file IO errors correctly.

        @param error_type: type of error to emulate,
                           'open' - patch open() to fail
                           'write' - patch write() to fail
        @param quiet: the value of 'quiet' argument for _makeInfoFiles()
        """
        # patch os.path.exists() to simulate that 'info' directory exists
        # but not 'admin' or 'host' files
        self.patch(os.path, "exists", lambda path: path.endswith("info"))

        # set-up requested IO error
        if error_type == "open":
            self.setUpOpenError()
        elif error_type == "write":
            self.setUpWriteError()
        else:
            self.fail(f"unexpected error_type '{error_type}'")

        # call _makeInfoFiles() and check that correct exception is raised
        with self.assertRaisesRegex(
            create_worker.CreateWorkerError,
            "could not write {}".format(_regexp_path("bdir", "info", "admin")),
        ):
            create_worker._makeInfoFiles("bdir", quiet)

        # check output to stdout
        if quiet:
            self.assertWasQuiet()
        else:
            self.assertStdoutEqual(
                "Creating {}, you need to edit it appropriately.\n".format(
                    os.path.join("info", "admin")
                )
            )

    def testOpenError(self) -> None:
        """
        test _makeInfoFiles() when open() fails
        """
        self.checkIOError("open", False)

    def testOpenErrorQuiet(self) -> None:
        """
        test _makeInfoFiles() when open() fails and quiet flag is enabled
        """
        self.checkIOError("open", True)

    def testWriteError(self) -> None:
        """
        test _makeInfoFiles() when write() fails
        """
        self.checkIOError("write", False)

    def testWriteErrorQuiet(self) -> None:
        """
        test _makeInfoFiles() when write() fails and quiet flag is enabled
        """
        self.checkIOError("write", True)

    def checkCreatedSuccessfully(self, quiet: bool) -> None:
        """
        Utility function to test _makeInfoFiles() when called on
        base directory that does not have 'info' sub-directory.

        @param quiet: the value of 'quiet' argument for _makeInfoFiles()
        """
        # patch os.path.exists() to report the no dirs/files exists
        self.patch(os.path, "exists", mock.Mock(return_value=False))
        # patch os.mkdir() to do nothing
        mkdir_mock = mock.Mock()
        self.patch(os, "mkdir", mkdir_mock)
        # capture calls to open() and write()
        self.setUpOpen()

        # call _makeInfoFiles()
        create_worker._makeInfoFiles("bdir", quiet)

        # check calls to os.mkdir()
        info_path = os.path.join("bdir", "info")
        mkdir_mock.assert_called_once_with(info_path)

        # check open() calls
        self.open.assert_has_calls([
            mock.call(os.path.join(info_path, "admin"), "w"),
            mock.call(os.path.join(info_path, "host"), "w"),
        ])

        # check write() calls
        self.fileobj.write.assert_has_calls([
            mock.call("Your Name Here <admin@youraddress.invalid>\n"),
            mock.call("Please put a description of this build host here\n"),
        ])

        # check output to stdout
        if quiet:
            self.assertWasQuiet()
        else:
            self.assertStdoutEqual(
                (
                    "mkdir {}\n"
                    "Creating {}, you need to edit it appropriately.\n"
                    "Creating {}, you need to edit it appropriately.\n"
                    "Not creating {} - add it if you wish\n"
                    "Please edit the files in {} appropriately.\n"
                ).format(
                    info_path,
                    os.path.join("info", "admin"),
                    os.path.join("info", "host"),
                    os.path.join("info", "access_uri"),
                    info_path,
                )
            )

    def testCreatedSuccessfully(self) -> None:
        """
        test calling _makeInfoFiles() on basedir without 'info' directory
        """
        self.checkCreatedSuccessfully(False)

    def testCreatedSuccessfullyQuiet(self) -> None:
        """
        test calling _makeInfoFiles() on basedir without 'info' directory
        and quiet flag is enabled
        """
        self.checkCreatedSuccessfully(True)

    def testInfoDirExists(self) -> None:
        """
        test calling _makeInfoFiles() on basedir with fully populated
        'info' directory
        """
        self.patch(os.path, "exists", mock.Mock(return_value=True))

        create_worker._makeInfoFiles("bdir", False)

        # there should be no messages to stdout
        self.assertWasQuiet()


class TestCreateWorker(misc.StdoutAssertionsMixin, TestDefaultOptionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.create_worker.createWorker()
    """

    def setUp(self) -> None:
        # capture stdout
        self.setUpStdoutAssertions()

    def setUpMakeFunctions(self, exception: BaseException | None = None) -> None:
        """
        patch create_worker._make*() functions with a mocks

        @param exception: if not None, the mocks will raise this exception.
        """
        self._makeBaseDir = mock.Mock(side_effect=exception)
        self.patch(create_worker, "_makeBaseDir", self._makeBaseDir)

        self._makeBuildbotTac = mock.Mock(side_effect=exception)
        self.patch(create_worker, "_makeBuildbotTac", self._makeBuildbotTac)

        self._makeInfoFiles = mock.Mock(side_effect=exception)
        self.patch(create_worker, "_makeInfoFiles", self._makeInfoFiles)

    def assertMakeFunctionsCalls(self, basedir: str, tac_contents: str, quiet: bool) -> None:
        """
        assert that create_worker._make*() were called with specified arguments
        """
        self._makeBaseDir.assert_called_once_with(basedir, quiet)
        self._makeBuildbotTac.assert_called_once_with(basedir, tac_contents, quiet)
        self._makeInfoFiles.assert_called_once_with(basedir, quiet)

    def testCreateError(self) -> None:
        """
        test that errors while creating worker directory are handled
        correctly by createWorker()
        """
        # patch _make*() functions to raise an exception
        self.setUpMakeFunctions(create_worker.CreateWorkerError("err-msg"))

        # call createWorker() and check that we get error exit code
        self.assertEqual(create_worker.createWorker(self.options), 1, "unexpected exit code")

        # check that correct error message was printed on stdout
        self.assertStdoutEqual("err-msg\nfailed to configure worker in bdir\n")

    def testMinArgs(self) -> None:
        """
        test calling createWorker() with only required arguments
        """
        # patch _make*() functions to do nothing
        self.setUpMakeFunctions()

        # call createWorker() and check that we get success exit code
        self.assertEqual(create_worker.createWorker(self.options), 0, "unexpected exit code")

        # check _make*() functions were called with correct arguments
        expected_tac_contents = create_worker._make_tac(self.options.copy())
        self.assertMakeFunctionsCalls(
            self.options["basedir"], expected_tac_contents, self.options["quiet"]
        )

        # check that correct info message was printed
        self.assertStdoutEqual("worker configured in bdir\n")

    def testQuiet(self) -> None:
        """
        test calling createWorker() with --quiet flag
        """
        options = self.options.copy()
        options["quiet"] = True

        # patch _make*() functions to do nothing
        self.setUpMakeFunctions()

        # call createWorker() and check that we get success exit code
        self.assertEqual(create_worker.createWorker(options), 0, "unexpected exit code")

        # check _make*() functions were called with correct arguments
        expected_tac_contents = create_worker._make_tac(self.options)
        self.assertMakeFunctionsCalls(options["basedir"], expected_tac_contents, options["quiet"])

        # there should be no output on stdout
        self.assertWasQuiet()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

from twisted.internet import threads
from twisted.python import threadpool
from twisted.trial.unittest import TestCase

from buildbot_worker.test.fake.reactor import NonThreadPool
from buildbot_worker.test.fake.reactor import TestReactor

if TYPE_CHECKING:
    from typing import Callable
    from typing import TypeVar

    from twisted.internet.defer import Deferred
    from twisted.internet.interfaces import IReactorFromThreads
    from twisted.python.threadpool import ThreadPool
    from typing_extensions import ParamSpec

    _T = TypeVar('_T')
    _P = ParamSpec('_P')


class TestReactorMixin:
    """
    Mix this in to get TestReactor as self.reactor which is correctly cleaned up
    at the end
    """

    def setup_test_reactor(self) -> None:
        assert isinstance(self, TestCase)
        self.patch(threadpool, 'ThreadPool', NonThreadPool)
        self.reactor = TestReactor()

        def deferToThread(
            f: Callable[_P, _T],
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> Deferred[_T]:
            return threads.deferToThreadPool(
                cast("IReactorFromThreads", self.reactor),
                cast("ThreadPool", self.reactor.getThreadPool()),
                f,
                *args,
                **kwargs,
            )

        self.patch(threads, 'deferToThread', deferToThread)

        self.addCleanup(self.reactor.stop)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import time

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.schedulers import timed
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler


class NightlyBase(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):
    """detailed getNextBuildTime tests"""

    OBJECTID = 133
    SCHEDULERID = 33

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    def makeScheduler(self, firstBuildDuration=0, **kwargs):
        return self.attachScheduler(timed.NightlyBase(**kwargs), self.OBJECTID, self.SCHEDULERID)

    @defer.inlineCallbacks
    def do_getNextBuildTime_test(self, sched, *expectations):
        for lastActuated, expected in expectations:
            # convert from tuples to epoch time (in local timezone)
            lastActuated_ep, expected_ep = [
                time.mktime(t + (0,) * (8 - len(t)) + (-1,)) for t in (lastActuated, expected)
            ]
            got_ep = yield sched.getNextBuildTime(lastActuated_ep)
            self.assertEqual(
                got_ep, expected_ep, f"{lastActuated} -> {time.localtime(got_ep)} != {expected}"
            )

    @defer.inlineCallbacks
    def test_getNextBuildTime_hourly(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'])
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 0, 0), (2011, 1, 1, 4, 0, 0)),
            ((2011, 1, 1, 3, 15, 0), (2011, 1, 1, 4, 0, 0)),
            ((2011, 1, 1, 3, 15, 1), (2011, 1, 1, 4, 0, 0)),
            ((2011, 1, 1, 3, 59, 1), (2011, 1, 1, 4, 0, 0)),
            ((2011, 1, 1, 3, 59, 59), (2011, 1, 1, 4, 0, 0)),
            ((2011, 1, 1, 23, 22, 22), (2011, 1, 2, 0, 0, 0)),
            ((2011, 1, 1, 23, 59, 0), (2011, 1, 2, 0, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_minutes_single(self):
        # basically the same as .._hourly
        sched = yield self.makeScheduler(name='test', builderNames=['test'], minute=4)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 0, 0), (2011, 1, 1, 3, 4, 0)),
            ((2011, 1, 1, 3, 15, 0), (2011, 1, 1, 4, 4, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_minutes_multiple(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], minute=[4, 34])
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 0, 0), (2011, 1, 1, 3, 4, 0)),
            ((2011, 1, 1, 3, 15, 0), (2011, 1, 1, 3, 34, 0)),
            ((2011, 1, 1, 3, 34, 0), (2011, 1, 1, 4, 4, 0)),
            ((2011, 1, 1, 3, 59, 1), (2011, 1, 1, 4, 4, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_minutes_star(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], minute='*')
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 11, 30), (2011, 1, 1, 3, 12, 0)),
            ((2011, 1, 1, 3, 12, 0), (2011, 1, 1, 3, 13, 0)),
            ((2011, 1, 1, 3, 59, 0), (2011, 1, 1, 4, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_hours_single(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], hour=4)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 0), (2011, 1, 1, 4, 0)),
            ((2011, 1, 1, 13, 0), (2011, 1, 2, 4, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_hours_multiple(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], hour=[7, 19])
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 0), (2011, 1, 1, 7, 0)),
            ((2011, 1, 1, 7, 1), (2011, 1, 1, 19, 0)),
            ((2011, 1, 1, 18, 59), (2011, 1, 1, 19, 0)),
            ((2011, 1, 1, 19, 59), (2011, 1, 2, 7, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_hours_minutes(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], hour=13, minute=19)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 1, 3, 11), (2011, 1, 1, 13, 19)),
            ((2011, 1, 1, 13, 19), (2011, 1, 2, 13, 19)),
            ((2011, 1, 1, 23, 59), (2011, 1, 2, 13, 19)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_month_single(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], month=3)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 2, 27, 3, 11), (2011, 3, 1, 0, 0)),
            # still hourly!
            ((2011, 3, 1, 1, 11), (2011, 3, 1, 2, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_month_multiple(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], month=[4, 6])
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 3, 30, 3, 11), (2011, 4, 1, 0, 0)),
            # still hourly!
            ((2011, 4, 1, 1, 11), (2011, 4, 1, 2, 0)),
            ((2011, 5, 29, 3, 11), (2011, 6, 1, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_month_dayOfMonth(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], month=[3, 6], dayOfMonth=[15]
        )
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 2, 12, 3, 11), (2011, 3, 15, 0, 0)),
            ((2011, 3, 12, 3, 11), (2011, 3, 15, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfMonth_single(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], dayOfMonth=10)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 9, 3, 0), (2011, 1, 10, 0, 0)),
            # still hourly!
            ((2011, 1, 10, 3, 0), (2011, 1, 10, 4, 0)),
            ((2011, 1, 30, 3, 0), (2011, 2, 10, 0, 0)),
            ((2011, 12, 30, 11, 0), (2012, 1, 10, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfMonth_multiple(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfMonth=[10, 20, 30]
        )
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 9, 22, 0), (2011, 1, 10, 0, 0)),
            ((2011, 1, 19, 22, 0), (2011, 1, 20, 0, 0)),
            ((2011, 1, 29, 22, 0), (2011, 1, 30, 0, 0)),
            # no Feb 30!
            ((2011, 2, 29, 22, 0), (2011, 3, 10, 0, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfMonth_hours_minutes(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfMonth=15, hour=20, minute=30
        )
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 13, 22, 19), (2011, 1, 15, 20, 30)),
            ((2011, 1, 15, 19, 19), (2011, 1, 15, 20, 30)),
            ((2011, 1, 15, 20, 29), (2011, 1, 15, 20, 30)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfWeek_single(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfWeek=1
        )  # Tuesday (2011-1-1 was a Saturday)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 3, 22, 19), (2011, 1, 4, 0, 0)),
            # still hourly!
            ((2011, 1, 4, 19, 19), (2011, 1, 4, 20, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfWeek_single_as_string(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfWeek="1"
        )  # Tuesday (2011-1-1 was a Saturday)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 3, 22, 19), (2011, 1, 4, 0, 0)),
            # still hourly!
            ((2011, 1, 4, 19, 19), (2011, 1, 4, 20, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfWeek_multiple_as_string(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfWeek="tue,3"
        )  # Tuesday, Thursday (2011-1-1 was a Saturday)
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 3, 22, 19), (2011, 1, 4, 0, 0)),
            # still hourly!
            ((2011, 1, 4, 19, 19), (2011, 1, 4, 20, 0)),
            ((2011, 1, 5, 22, 19), (2011, 1, 6, 0, 0)),
            # still hourly!
            ((2011, 1, 6, 19, 19), (2011, 1, 6, 20, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfWeek_multiple_hours(self):
        # Tuesday, Thursday (2011-1-1 was a Saturday)
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfWeek=[1, 3], hour=1
        )
        yield self.master.startService()

        yield self.do_getNextBuildTime_test(
            sched,
            ((2011, 1, 3, 22, 19), (2011, 1, 4, 1, 0)),
            ((2011, 1, 4, 22, 19), (2011, 1, 6, 1, 0)),
        )

    @defer.inlineCallbacks
    def test_getNextBuildTime_dayOfWeek_dayOfMonth(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], dayOfWeek=[1, 4], dayOfMonth=5, hour=1
        )
        yield self.master.startService()
        yield self.do_getNextBuildTime_test(
            sched,
            # Tues
            ((2011, 1, 3, 22, 19), (2011, 1, 4, 1, 0)),
            # 5th
            ((2011, 1, 4, 22, 19), (2011, 1, 5, 1, 0)),
            # Thurs
            ((2011, 1, 5, 22, 19), (2011, 1, 7, 1, 0)),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json
import os
import webbrowser
from unittest import mock

import twisted
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet import threads
from twisted.python import failure
from twisted.trial import unittest
from twisted.web.resource import Resource

import buildbot
from buildbot.process.properties import Secret
from buildbot.secrets.manager import SecretManager
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.site import SiteWithClose
from buildbot.util import bytes2unicode

try:
    import requests
except ImportError:
    requests = None  # type: ignore[assignment]


if requests:
    from buildbot.www import oauth2  # pylint: disable=ungrouped-imports
else:
    oauth2 = None  # type: ignore[assignment]


class FakeResponse:
    def __init__(self, _json):
        self.json = lambda: _json
        self.content = json.dumps(_json)

    def raise_for_status(self):
        pass


class OAuth2Auth(TestReactorMixin, www.WwwTestMixin, ConfigErrorsMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        if requests is None:
            raise unittest.SkipTest("Need to install requests to test oauth2")

        self.patch(requests, 'request', mock.Mock(spec=requests.request))
        self.patch(requests, 'post', mock.Mock(spec=requests.post))
        self.patch(requests, 'get', mock.Mock(spec=requests.get))

    @defer.inlineCallbacks
    def setup_google_auth(self):
        auth = oauth2.GoogleAuth("ggclientID", "clientSECRET")
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth(self):
        auth = oauth2.GitHubAuth("ghclientID", "clientSECRET")
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth_v4(self):
        auth = oauth2.GitHubAuth("ghclientID", "clientSECRET", apiVersion=4)
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth_v4_secret(self):
        auth = oauth2.GitHubAuth(Secret("client-id"), Secret("client-secret"), apiVersion=4)
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        fake_storage_service = FakeSecretStorage()
        fake_storage_service.reconfigService(
            secretdict={"client-id": "secretClientId", "client-secret": "secretClientSecret"}
        )
        secret_service = SecretManager()
        secret_service.services = [fake_storage_service]
        yield secret_service.setServiceParent(master)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth_v4_teams(self):
        auth = oauth2.GitHubAuth(
            "ghclientID", "clientSECRET", apiVersion=4, getTeamsMembership=True
        )
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth_enterprise(self):
        auth = oauth2.GitHubAuth(
            "ghclientID", "clientSECRET", serverURL="https://git.corp.fakecorp.com"
        )
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_github_auth_enterprise_v4(self):
        auth = oauth2.GitHubAuth(
            "ghclientID",
            "clientSECRET",
            apiVersion=4,
            getTeamsMembership=True,
            serverURL="https://git.corp.fakecorp.com",
        )
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_gitlab_auth(self):
        auth = oauth2.GitLabAuth("https://gitlab.test/", "glclientID", "clientSECRET")
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def setup_bitbucket_auth(self):
        auth = oauth2.BitbucketAuth("bbclientID", "clientSECRET")
        master = yield self.make_master(url='h:/a/b/', auth=auth)
        auth.reconfigAuth(master, master.config)
        return auth

    @defer.inlineCallbacks
    def test_getGoogleLoginURL(self):
        auth = yield self.setup_google_auth()
        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://accounts.google.com/o/oauth2/auth?client_id=ggclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.email+"
            "https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.profile&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://accounts.google.com/o/oauth2/auth?client_id=ggclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.email+"
            "https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.profile"
        )

        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getGithubLoginURL(self):
        auth = yield self.setup_github_auth()
        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://github.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://github.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getGithubLoginURL_with_secret(self):
        auth = yield self.setup_github_auth_v4_secret()
        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://github.com/login/oauth/authorize?client_id=secretClientId&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://github.com/login/oauth/authorize?client_id=secretClientId&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getGithubELoginURL(self):
        auth = yield self.setup_github_auth_enterprise()

        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://git.corp.fakecorp.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://git.corp.fakecorp.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getGithubLoginURL_v4(self):
        auth = yield self.setup_github_auth_enterprise_v4()

        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://git.corp.fakecorp.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://git.corp.fakecorp.com/login/oauth/authorize?client_id=ghclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&response_type=code&"
            "scope=user%3Aemail+read%3Aorg"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getGitLabLoginURL(self):
        auth = yield self.setup_gitlab_auth()

        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://gitlab.test/oauth/authorize"
            "?client_id=glclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&"
            "response_type=code&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://gitlab.test/oauth/authorize"
            "?client_id=glclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&"
            "response_type=code"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_getBitbucketLoginURL(self):
        auth = yield self.setup_bitbucket_auth()

        res = yield auth.getLoginURL('http://redir')
        exp = (
            "https://bitbucket.org/site/oauth2/authorize?"
            "client_id=bbclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&"
            "response_type=code&"
            "state=redirect%3Dhttp%253A%252F%252Fredir"
        )
        self.assertEqual(res, exp)
        res = yield auth.getLoginURL(None)
        exp = (
            "https://bitbucket.org/site/oauth2/authorize?"
            "client_id=bbclientID&"
            "redirect_uri=h%3A%2Fa%2Fb%2Fauth%2Flogin&"
            "response_type=code"
        )
        self.assertEqual(res, exp)

    @defer.inlineCallbacks
    def test_GoogleVerifyCode(self):
        auth = yield self.setup_google_auth()

        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]
        auth.get = mock.Mock(
            side_effect=[{"name": 'foo bar', "email": 'bar@foo', "picture": 'http://pic'}]
        )
        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                'avatar_url': 'http://pic',
                'email': 'bar@foo',
                'full_name': 'foo bar',
                'username': 'bar',
            },
            res,
        )

    @defer.inlineCallbacks
    def test_GithubVerifyCode(self):
        auth = yield self.setup_github_auth()

        test = self
        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]

        def fake_get(self, ep, **kwargs):
            test.assertEqual(
                self.headers,
                {
                    'Authorization': 'token TOK3N',
                    'User-Agent': f'buildbot/{buildbot.version}',
                },
            )
            if ep == '/user':
                return {"login": 'bar', "name": 'foo bar', "email": 'buzz@bar'}
            if ep == '/user/emails':
                return [
                    {'email': 'buzz@bar', 'verified': True, 'primary': False},
                    {'email': 'bar@foo', 'verified': True, 'primary': True},
                ]
            if ep == '/user/orgs':
                return [
                    {"login": 'hello'},
                    {"login": 'grp'},
                ]
            return None

        auth.get = fake_get

        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                'email': 'bar@foo',
                'username': 'bar',
                'groups': ["hello", "grp"],
                'full_name': 'foo bar',
            },
            res,
        )

    @defer.inlineCallbacks
    def test_GithubVerifyCode_v4(self):
        auth = yield self.setup_github_auth_v4()

        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]
        auth.post = mock.Mock(
            side_effect=[
                {
                    'data': {
                        'viewer': {
                            'organizations': {
                                'edges': [{'node': {'login': 'hello'}}, {'node': {'login': 'grp'}}]
                            },
                            'login': 'bar',
                            'email': 'bar@foo',
                            'name': 'foo bar',
                        }
                    }
                }
            ]
        )
        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                'email': 'bar@foo',
                'username': 'bar',
                'groups': ["hello", "grp"],
                'full_name': 'foo bar',
            },
            res,
        )

    @defer.inlineCallbacks
    def test_GithubVerifyCode_v4_teams(self):
        auth = yield self.setup_github_auth_v4_teams()

        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]
        auth.post = mock.Mock(
            side_effect=[
                {
                    'data': {
                        'viewer': {
                            'organizations': {
                                'edges': [{'node': {'login': 'hello'}}, {'node': {'login': 'grp'}}]
                            },
                            'login': 'bar',
                            'email': 'bar@foo',
                            'name': 'foo bar',
                        }
                    }
                },
                {
                    'data': {
                        'hello': {
                            'teams': {
                                'edges': [
                                    {'node': {'name': 'developers', 'slug': 'develpers'}},
                                    {'node': {'name': 'contributors', 'slug': 'contributors'}},
                                ]
                            }
                        },
                        'grp': {
                            'teams': {
                                'edges': [
                                    {'node': {'name': 'developers', 'slug': 'develpers'}},
                                    {'node': {'name': 'contributors', 'slug': 'contributors'}},
                                    {'node': {'name': 'committers', 'slug': 'committers'}},
                                    {
                                        'node': {
                                            'name': 'Team with spaces and caps',
                                            'slug': 'team-with-spaces-and-caps',
                                        }
                                    },
                                ]
                            }
                        },
                    }
                },
            ]
        )
        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                'email': 'bar@foo',
                'username': 'bar',
                'groups': [
                    'hello',
                    'grp',
                    'grp/Team with spaces and caps',
                    'grp/committers',
                    'grp/contributors',
                    'grp/developers',
                    'grp/develpers',
                    'grp/team-with-spaces-and-caps',
                    'hello/contributors',
                    'hello/developers',
                    'hello/develpers',
                ],
                'full_name': 'foo bar',
            },
            res,
        )

    def test_GitHubAuthBadApiVersion(self):
        for bad_api_version in (2, 5, 'a'):
            with self.assertRaisesConfigError('GitHubAuth apiVersion must be 3 or 4 not '):
                oauth2.GitHubAuth("ghclientID", "clientSECRET", apiVersion=bad_api_version)

    def test_GitHubAuthRaiseErrorWithApiV3AndGetTeamMembership(self):
        with self.assertRaisesConfigError(
            'Retrieving team membership information using '
            'GitHubAuth is only possible using GitHub api v4.'
        ):
            oauth2.GitHubAuth("ghclientID", "clientSECRET", apiVersion=3, getTeamsMembership=True)

    @defer.inlineCallbacks
    def test_GitlabVerifyCode(self):
        auth = yield self.setup_gitlab_auth()

        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]
        auth.get = mock.Mock(
            side_effect=[
                {  # /user
                    "name": "Foo Bar",
                    "username": "fbar",
                    "id": 5,
                    "avatar_url": "https://avatar/fbar.png",
                    "email": "foo@bar",
                    "twitter": "fb",
                },
                [  # /groups
                    {"id": 10, "name": "Hello", "path": "hello"},
                    {"id": 20, "name": "Group", "path": "grp"},
                ],
            ]
        )
        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                "full_name": "Foo Bar",
                "username": "fbar",
                "email": "foo@bar",
                "avatar_url": "https://avatar/fbar.png",
                "groups": ["hello", "grp"],
            },
            res,
        )

    @defer.inlineCallbacks
    def test_BitbucketVerifyCode(self):
        auth = yield self.setup_bitbucket_auth()

        requests.get.side_effect = []
        requests.post.side_effect = [FakeResponse({"access_token": 'TOK3N'})]
        auth.get = mock.Mock(
            side_effect=[
                {"username": 'bar', "display_name": 'foo bar'},  # /user
                {
                    "values": [
                        {'email': 'buzz@bar', 'is_primary': False},
                        {'email': 'bar@foo', 'is_primary': True},
                    ]
                },  # /user/emails
                {"values": [{'slug': 'hello'}, {'slug': 'grp'}]},  # /workspaces?role=member
            ]
        )
        res = yield auth.verifyCode("code!")
        self.assertEqual(
            {
                'email': 'bar@foo',
                'username': 'bar',
                "groups": ["hello", "grp"],
                'full_name': 'foo bar',
            },
            res,
        )

    @defer.inlineCallbacks
    def test_loginResource(self):
        auth = yield self.setup_github_auth()

        class fakeAuth:
            homeUri = "://me"
            getLoginURL = mock.Mock(side_effect=lambda x: defer.succeed("://"))
            verifyCode = mock.Mock(side_effect=lambda code: defer.succeed({"username": "bar"}))
            acceptToken = mock.Mock(side_effect=lambda token: defer.succeed({"username": "bar"}))
            userInfoProvider = None

        rsrc = auth.getLoginResource()
        rsrc.auth = fakeAuth()
        res = yield self.render_resource(rsrc, b'/')
        rsrc.auth.getLoginURL.assert_called_once_with(None)
        rsrc.auth.verifyCode.assert_not_called()
        self.assertEqual(res, {'redirected': b'://'})
        rsrc.auth.getLoginURL.reset_mock()
        rsrc.auth.verifyCode.reset_mock()
        res = yield self.render_resource(rsrc, b'/?code=code!')
        rsrc.auth.getLoginURL.assert_not_called()
        rsrc.auth.verifyCode.assert_called_once_with(b"code!")
        self.assertEqual(self.master.session.user_info, {'username': 'bar'})
        self.assertEqual(res, {'redirected': b'://me'})
        # token not supported anymore
        res = yield self.render_resource(rsrc, b'/?token=token!')
        rsrc.auth.getLoginURL.assert_called_once()

    @defer.inlineCallbacks
    def test_getConfig_github(self):
        auth = yield self.setup_github_auth()
        self.assertEqual(
            auth.getConfigDict(),
            {'fa_icon': 'fa-github', 'autologin': False, 'name': 'GitHub', 'oauth2': True},
        )

    @defer.inlineCallbacks
    def test_getConfig_google(self):
        auth = yield self.setup_google_auth()
        self.assertEqual(
            auth.getConfigDict(),
            {'fa_icon': 'fa-google-plus', 'autologin': False, 'name': 'Google', 'oauth2': True},
        )

    @defer.inlineCallbacks
    def test_getConfig_gitlab(self):
        auth = yield self.setup_gitlab_auth()
        self.assertEqual(
            auth.getConfigDict(),
            {'fa_icon': 'fa-git', 'autologin': False, 'name': 'GitLab', 'oauth2': True},
        )

    @defer.inlineCallbacks
    def test_getConfig_bitbucket(self):
        auth = yield self.setup_bitbucket_auth()
        self.assertEqual(
            auth.getConfigDict(),
            {'fa_icon': 'fa-bitbucket', 'autologin': False, 'name': 'Bitbucket', 'oauth2': True},
        )


# unit tests are not very useful to write new oauth support
# so following is an e2e test, which opens a browser, and do the oauth
# negotiation. The browser window close in the end of the test

# in order to use this tests, you need to create Github/Google ClientID (see doc on how to do it)
# point OAUTHCONF environment variable to a file with following params:
#  {
#  "GitHubAuth": {
#     "CLIENTID": "XX
#     "CLIENTSECRET": "XX"
#  },
#  "GoogleAuth": {
#     "CLIENTID": "XX",
#     "CLIENTSECRET": "XX"
#  }
#  "GitLabAuth": {
#     "INSTANCEURI": "XX",
#     "CLIENTID": "XX",
#     "CLIENTSECRET": "XX"
#  }
#  }


class OAuth2AuthGitHubE2E(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    authClass = "GitHubAuth"

    def _instantiateAuth(self, cls, config):
        return cls(config["CLIENTID"], config["CLIENTSECRET"])

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        if requests is None:
            raise unittest.SkipTest("Need to install requests to test oauth2")

        if "OAUTHCONF" not in os.environ:
            raise unittest.SkipTest(
                "Need to pass OAUTHCONF path to json file via environ to run this e2e test"
            )

        with open(os.environ['OAUTHCONF'], encoding='utf-8') as f:
            jsonData = f.read()
        config = json.loads(jsonData)[self.authClass]
        from buildbot.www import oauth2

        self.auth = self._instantiateAuth(getattr(oauth2, self.authClass), config)

        # 5000 has to be hardcoded, has oauth clientids are bound to a fully
        # classified web site
        master = yield self.make_master(url='http://localhost:5000/', auth=self.auth)
        self.auth.reconfigAuth(master, master.config)

    def tearDown(self):
        from twisted.internet.tcp import Server

        # browsers has the bad habit on not closing the persistent
        # connections, so we need to hack them away to make trial happy
        f = failure.Failure(Exception("test end"))
        for reader in reactor.getReaders():
            if isinstance(reader, Server):
                reader.connectionLost(f)

    @defer.inlineCallbacks
    def test_E2E(self):
        d = defer.Deferred()
        twisted.web.http._logDateTimeUsers = 1

        class HomePage(Resource):
            isLeaf = True

            def render_GET(self, request):
                info = request.getSession().user_info
                reactor.callLater(0, d.callback, info)
                return (
                    b"<html><script>setTimeout(close,1000)</script><body>WORKED: "
                    + info
                    + b"</body></html>"
                )

        class MySite(SiteWithClose):
            def makeSession(self):
                uid = self._mkuid()
                session = self.sessions[uid] = self.sessionFactory(self, uid)
                return session

        root = Resource()
        root.putChild(b"", HomePage())
        auth = Resource()
        root.putChild(b'auth', auth)
        auth.putChild(b'login', self.auth.getLoginResource())
        site = MySite(root)
        listener = reactor.listenTCP(5000, site)

        def thd():
            res = requests.get('http://localhost:5000/auth/login', timeout=30)
            content = bytes2unicode(res.content)
            webbrowser.open(content)

        threads.deferToThread(thd)
        res = yield d
        yield listener.stopListening()
        yield site.stopFactory()
        yield site.close_connections()

        self.assertIn("full_name", res)
        self.assertIn("email", res)
        self.assertIn("username", res)


class OAuth2AuthGoogleE2E(OAuth2AuthGitHubE2E):
    authClass = "GoogleAuth"


class OAuth2AuthGitLabE2E(OAuth2AuthGitHubE2E):
    authClass = "GitLabAuth"

    def _instantiateAuth(self, cls, config):
        return cls(config["INSTANCEURI"], config["CLIENTID"], config["CLIENTSECRET"])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import log
from twisted.trial import unittest

from buildbot.db.buildrequests import BuildRequestModel
from buildbot.db.buildsets import BuildSetModel
from buildbot.process import properties
from buildbot.schedulers import triggerable
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import interfaces
from buildbot.test.util import scheduler


class TriggerableInterfaceTest(unittest.TestCase, interfaces.InterfaceTests):
    def test_interface(self):
        self.assertInterfacesImplemented(triggerable.Triggerable)


class Triggerable(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):
    OBJECTID = 33
    SCHEDULERID = 13

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        # Necessary to get an assertable submitted_at time.
        self.reactor.advance(946684799)

        yield self.setUpScheduler()
        self.subscription = None

    @defer.inlineCallbacks
    def makeScheduler(self, overrideBuildsetMethods=False, **kwargs):
        yield self.master.db.insert_test_data([fakedb.Builder(id=77, name='b')])

        sched = yield self.attachScheduler(
            triggerable.Triggerable(name='n', builderNames=['b'], **kwargs),
            self.OBJECTID,
            self.SCHEDULERID,
            overrideBuildsetMethods=overrideBuildsetMethods,
        )

        return sched

    @defer.inlineCallbacks
    def assertTriggeredBuildset(self, idsDeferred, waited_for, properties=None, sourcestamps=None):
        if properties is None:
            properties = {}
        bsid, brids = yield idsDeferred
        properties.update({'scheduler': ('n', 'Scheduler')})

        buildset = yield self.master.db.buildsets.getBuildset(bsid)
        got_properties = yield self.master.db.buildsets.getBuildsetProperties(bsid)
        self.assertEqual(got_properties, properties)

        from datetime import datetime

        from buildbot.util import UTC

        self.assertEqual(
            buildset,
            BuildSetModel(
                bsid=bsid,
                external_idstring=None,
                reason="The Triggerable scheduler named 'n' triggered this build",
                submitted_at=datetime(1999, 12, 31, 23, 59, 59, tzinfo=UTC),
                results=-1,
                # sourcestamps testing is just after
                sourcestamps=buildset.sourcestamps,
            ),
        )

        actual_sourcestamps = yield defer.gatherResults(
            [self.master.db.sourcestamps.getSourceStamp(ssid) for ssid in buildset.sourcestamps],
            consumeErrors=True,
        )

        self.assertEqual(len(sourcestamps), len(actual_sourcestamps))
        for expected_ss, actual_ss in zip(sourcestamps, actual_sourcestamps):
            # We don't care if the actual sourcestamp has *more* attributes
            # than expected.
            self.assertEqual(expected_ss, {k: getattr(actual_ss, k) for k in expected_ss.keys()})

        for brid in brids.values():
            buildrequest = yield self.master.db.buildrequests.getBuildRequest(brid)
            self.assertEqual(
                buildrequest,
                BuildRequestModel(
                    buildrequestid=brid,
                    buildername='b',
                    builderid=77,
                    buildsetid=bsid,
                    claimed_at=None,
                    complete=False,
                    complete_at=None,
                    claimed_by_masterid=None,
                    priority=0,
                    results=-1,
                    submitted_at=datetime(1999, 12, 31, 23, 59, 59, tzinfo=UTC),
                    waited_for=waited_for,
                ),
            )

        return bsid

    def sendCompletionMessage(self, bsid, results=3):
        self.master.mq.callConsumer(
            ('buildsets', str(bsid), 'complete'),
            {
                "bsid": bsid,
                "submitted_at": 100,
                "complete": True,
                "complete_at": 200,
                "external_idstring": None,
                "reason": 'triggering',
                "results": results,
                "sourcestamps": [],
                "parent_buildid": None,
                "parent_relationship": None,
            },
        )

    # tests

    # NOTE: these tests take advantage of the fact that all of the fake
    # scheduler operations are synchronous, and thus do not return a Deferred.
    # The Deferred from trigger() is completely processed before this test
    # method returns.

    @defer.inlineCallbacks
    def test_constructor_no_reason(self):
        sched = yield self.makeScheduler()
        yield sched.configureService()
        self.assertEqual(sched.reason, None)  # default reason is dynamic

    @defer.inlineCallbacks
    def test_constructor_explicit_reason(self):
        sched = yield self.makeScheduler(reason="Because I said so")
        yield sched.configureService()
        self.assertEqual(sched.reason, "Because I said so")

    @defer.inlineCallbacks
    def test_constructor_priority_none(self):
        sched = yield self.makeScheduler(priority=None)
        yield sched.configureService()
        self.assertEqual(sched.priority, None)

    @defer.inlineCallbacks
    def test_constructor_priority_int(self):
        sched = yield self.makeScheduler(priority=8)
        yield sched.configureService()
        self.assertEqual(sched.priority, 8)

    @defer.inlineCallbacks
    def test_constructor_priority_function(self):
        def sched_priority(builderNames, changesByCodebase):
            return 0

        sched = yield self.makeScheduler(priority=sched_priority)
        yield sched.configureService()
        self.assertEqual(sched.priority, sched_priority)

    @defer.inlineCallbacks
    def test_trigger(self):
        sched = yield self.makeScheduler(codebases={'cb': {'repository': 'r'}})

        # no subscription should be in place yet
        self.assertEqual(sched.master.mq.qrefs, [])

        yield self.master.startService()

        # trigger the scheduler, exercising properties while we're at it
        waited_for = True
        set_props = properties.Properties()
        set_props.setProperty('pr', 'op', 'test')
        ss = {
            'revision': 'myrev',
            'branch': 'br',
            'project': 'p',
            'repository': 'r',
            'codebase': 'cb',
        }
        idsDeferred, d = sched.trigger(waited_for, sourcestamps=[ss], set_props=set_props)
        self.reactor.advance(0)  # let the debounced function fire

        bsid = yield self.assertTriggeredBuildset(
            idsDeferred,
            waited_for,
            properties={'pr': ('op', 'test')},
            sourcestamps=[
                {
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "codebase": 'cb',
                    "revision": 'myrev',
                },
            ],
        )

        # set up a boolean so that we can know when the deferred fires
        self.fired = False

        @d.addCallback
        def fired(xxx_todo_changeme):
            (result, brids) = xxx_todo_changeme
            self.assertEqual(result, 3)  # from sendCompletionMessage
            self.assertEqual(brids, {77: 1})
            self.fired = True

        d.addErrback(log.err)

        # check that the scheduler has subscribed to buildset changes, but
        # not fired yet
        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs],
            [
                ('schedulers', '13', 'updated'),
                ('buildsets', None, 'complete'),
            ],
        )
        self.assertFalse(self.fired)

        # pretend a non-matching buildset is complete
        self.sendCompletionMessage(27)

        # scheduler should not have reacted
        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs],
            [
                ('schedulers', '13', 'updated'),
                ('buildsets', None, 'complete'),
            ],
        )
        self.assertFalse(self.fired)

        # pretend the matching buildset is complete
        self.sendCompletionMessage(bsid)
        self.reactor.advance(0)  # let the debounced function fire

        # scheduler should have reacted
        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs], [('schedulers', '13', 'updated')]
        )
        self.assertTrue(self.fired)
        yield d

    @defer.inlineCallbacks
    def test_trigger_overlapping(self):
        sched = yield self.makeScheduler(codebases={'cb': {'repository': 'r'}})

        # no subscription should be in place yet
        self.assertEqual(sched.master.mq.qrefs, [])

        yield self.master.startService()

        waited_for = False

        def makeSS(rev):
            return {
                'revision': rev,
                'branch': 'br',
                'project': 'p',
                'repository': 'r',
                'codebase': 'cb',
            }

        # trigger the scheduler the first time
        idsDeferred, d = sched.trigger(waited_for, [makeSS('myrev1')])  # triggers bsid 200
        bsid1 = yield self.assertTriggeredBuildset(
            idsDeferred,
            waited_for,
            sourcestamps=[
                {
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "codebase": 'cb',
                    "revision": 'myrev1',
                },
            ],
        )
        d.addCallback(
            lambda res_brids: self.assertEqual(res_brids[0], 11)
            and self.assertEqual(res_brids[1], {77: 1})
        )

        waited_for = True
        # and the second time
        idsDeferred, d = sched.trigger(waited_for, [makeSS('myrev2')])  # triggers bsid 201
        self.reactor.advance(0)  # let the debounced function fire
        bsid2 = yield self.assertTriggeredBuildset(
            idsDeferred,
            waited_for,
            sourcestamps=[
                {
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "codebase": 'cb',
                    "revision": 'myrev2',
                },
            ],
        )
        d.addCallback(
            lambda res_brids1: self.assertEqual(res_brids1[0], 22)
            and self.assertEqual(res_brids1[1], {77: 2})
        )

        # check that the scheduler has subscribed to buildset changes
        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs],
            [
                ('schedulers', '13', 'updated'),
                ('buildsets', None, 'complete'),
            ],
        )

        # let a few buildsets complete
        self.sendCompletionMessage(29, results=3)
        self.sendCompletionMessage(bsid2, results=22)
        self.sendCompletionMessage(9, results=3)
        self.sendCompletionMessage(bsid1, results=11)
        self.reactor.advance(0)  # let the debounced function fire

        # both should have triggered with appropriate results, and the
        # subscription should be cancelled
        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs], [('schedulers', '13', 'updated')]
        )

    @defer.inlineCallbacks
    def test_trigger_with_sourcestamp(self):
        # Test triggering a scheduler with a sourcestamp, and see that
        # sourcestamp handed to addBuildsetForSourceStampsWithDefaults.
        sched = yield self.makeScheduler(overrideBuildsetMethods=True)
        yield self.master.startService()

        waited_for = False
        ss = {
            'repository': 'r3',
            'codebase': 'cb3',
            'revision': 'fixrev3',
            'branch': 'default',
            'project': 'p',
        }
        idsDeferred = sched.trigger(waited_for, sourcestamps=[ss])[0]
        yield idsDeferred

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': {'scheduler': ('n', 'Scheduler')},
                        'reason': "The Triggerable scheduler named 'n' triggered this build",
                        'sourcestamps': [
                            {
                                'branch': 'default',
                                'codebase': 'cb3',
                                'project': 'p',
                                'repository': 'r3',
                                'revision': 'fixrev3',
                            },
                        ],
                        'waited_for': False,
                    },
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_trigger_without_sourcestamps(self):
        # Test triggering *without* sourcestamps, and see that nothing is passed
        # to addBuildsetForSourceStampsWithDefaults
        waited_for = True
        sched = yield self.makeScheduler(overrideBuildsetMethods=True)
        yield self.master.startService()

        idsDeferred = sched.trigger(waited_for, sourcestamps=[])[0]
        yield idsDeferred

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': {'scheduler': ('n', 'Scheduler')},
                        'reason': "The Triggerable scheduler named 'n' triggered this build",
                        'sourcestamps': [],
                        'waited_for': True,
                    },
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_trigger_with_reason(self):
        # Test triggering with a reason, and make sure the buildset's reason is updated accordingly
        # (and not the default)
        waited_for = True
        sched = yield self.makeScheduler(overrideBuildsetMethods=True)
        yield self.master.startService()

        set_props = properties.Properties()
        set_props.setProperty('reason', 'test1', 'test')
        idsDeferred, _ = sched.trigger(waited_for, sourcestamps=[], set_props=set_props)
        yield idsDeferred

        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        'builderNames': None,
                        'priority': None,
                        'properties': {
                            'scheduler': ('n', 'Scheduler'),
                            'reason': ('test1', 'test'),
                        },
                        'reason': "test1",
                        'sourcestamps': [],
                        'waited_for': True,
                    },
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_startService_stopService(self):
        yield self.makeScheduler()
        yield self.master.startService()
        yield self.master.stopService()
